# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/053_optimizer.ipynb (unless otherwise specified).

__all__ = ['wrap_optimizer', 'adabelief', 'adamw']

# Cell
from .imports import *
from fastai.optimizer import *

# Cell
import torch_optimizer as optim

# Cell
def wrap_optimizer(opt, **kwargs): return partial(OptimWrapper, opt=opt, **kwargs)

# Cell
adabelief = wrap_optimizer(optim.AdaBelief)
adamw = wrap_optimizer(torch.optim.AdamW)